﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

using System.Text.RegularExpressions;

namespace howto_measure_map
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        // The loaded map.
        private Bitmap Map = null;

        // Known units.
        enum Units
        {
            Undefined,
            Miles,
            Yards,
            Feet,
            Kilometers,
            Meters,
        };

        // Key map values.
        private double ScaleDistanceInUnits = -1;
        private double ScaleDistanceInPixels = -1;
        private Units CurrentUnit = Units.Miles;
        private double CurrentDistance = 0;

        // Open a map image.
        private void mnuFileOpen_Click(object sender, EventArgs e)
        {
            if (ofdMap.ShowDialog() == DialogResult.OK)
            {
                Map = LoadUnlocked(ofdMap.FileName);
                picMap.Image = Map;
            }

            btnScale.Enabled = true;
            btnDistance.Enabled = true;
        }

        // Set the desired units.
        private void btnUnit_Click(object sender, EventArgs e)
        {
            Units old_unit = CurrentUnit;
            ToolStripMenuItem menu_item = sender as ToolStripMenuItem;
            switch (menu_item.Text)
            {
                case "Miles":
                    CurrentUnit = Units.Miles;
                    break;
                case "Yards":
                    CurrentUnit = Units.Yards;
                    break;
                case "Feet":
                    CurrentUnit = Units.Feet;
                    break;
                case "Kilometers":
                    CurrentUnit = Units.Kilometers;
                    break;
                case "Meters":
                    CurrentUnit = Units.Meters;
                    break;
            }
            btnUnits.Text = CurrentUnit.ToString();

            // Display the map scale and distance in this unit.
            // Find a factor to convert from the old units to meters.
            double conversion = 1;
            if (old_unit == Units.Feet) conversion = 0.3048;
            else if (old_unit == Units.Yards) conversion = 0.9144;
            else if (old_unit == Units.Miles) conversion = 1609.344;
            else if (old_unit == Units.Kilometers) conversion = 1000;

            // Find a factor to convert from meters to the new units.
            if (CurrentUnit == Units.Feet) conversion *= 3.28083;
            else if (CurrentUnit == Units.Yards) conversion *= 1.09361;
            else if (CurrentUnit == Units.Miles) conversion *= 0.000621;
            else if (CurrentUnit == Units.Kilometers) conversion *= 0.001;

            // Convert and display the values.
            ScaleDistanceInUnits *= conversion;
            CurrentDistance *= conversion;

            btnScale.Text = string.Format("{0} {1}/pixel",
                ScaleDistanceInUnits / ScaleDistanceInPixels,
                CurrentUnit.ToString());
            btnDistance.Text = string.Format("Distance: {0} {1}",
                CurrentDistance, CurrentUnit);
        }

        // Reset the scale.
        private Point StartPoint, EndPoint;
        private void btnScale_Click(object sender, EventArgs e)
        {
            lblInstructions.Text = "Click and drag from the start and end point of the map's scale bar.";
            picMap.Cursor = Cursors.Cross;

            picMap.MouseDown += Scale_MouseDown;
        }
        private void Scale_MouseDown(object sender, MouseEventArgs e)
        {
            StartPoint = e.Location;
            picMap.MouseDown -= Scale_MouseDown;
            picMap.MouseMove += Scale_MouseMove;
            picMap.MouseUp += Scale_MouseUp;
        }
        private void Scale_MouseMove(object sender, MouseEventArgs e)
        {
            EndPoint = e.Location;
            DisplayScaleLine();
        }
        private void Scale_MouseUp(object sender, MouseEventArgs e)
        {
            picMap.MouseMove -= Scale_MouseMove;
            picMap.MouseUp -= Scale_MouseUp;
            picMap.Cursor = Cursors.Default;
            lblInstructions.Text = "";

            // Get the scale.
            ScaleDialog dlg = new ScaleDialog();
            if (dlg.ShowDialog() == DialogResult.OK)
            {
                // Get the distance on the screen.
                int dx = EndPoint.X - StartPoint.X;
                int dy = EndPoint.Y - StartPoint.Y;
                double dist = Math.Sqrt(dx * dx + dy * dy);
                if (dist < 1) return;
                ScaleDistanceInPixels = dist;

                // Parse the distance.
                ParseDistanceString(dlg.txtScaleLength.Text,
                    out ScaleDistanceInUnits, out CurrentUnit);

                // Display the units.
                btnUnits.Text = CurrentUnit.ToString();

                // Display the scale.
                btnScale.Text = string.Format("{0} {1}/pixel",
                    ScaleDistanceInUnits / ScaleDistanceInPixels,
                    CurrentUnit.ToString());

                // Clear the distance.
                btnDistance.Text = "Distance <undefined>";
            }
        }

        // Load the image into a Bitmap, clone it, and
        // set the PictureBox's Image property to the Bitmap.
        private Bitmap LoadUnlocked(string file_name)
        {
            using (Bitmap bm = new Bitmap(file_name))
            {
                Bitmap result = new Bitmap(bm.Width, bm.Height);
                using (Graphics gr = Graphics.FromImage(result))
                {
                    gr.DrawImage(bm, 0, 0);
                }
                return result;
            }
        }

        // Display the scale line.
        private void DisplayScaleLine()
        {
            Bitmap bm = (Bitmap)Map.Clone();
            using (Graphics gr = Graphics.FromImage(bm))
            {
                gr.DrawLine(Pens.Red, StartPoint, EndPoint);
            }
            picMap.Image = bm;
        }

        // Parse a distance string. Return the length in meters.
        private void ParseDistanceString(string txt, out double distance, out Units unit)
        {
            txt = txt.Trim();

            // Find the longest substring that makes sense as a double.
            int i = DoublePrefixLength(txt);
            if (i <= 0)
            {
                distance = -1;
                unit = Units.Undefined;
            }
            else
            {
                // Get the distance.
                distance = double.Parse(txt.Substring(0, i));

                // Get the unit.
                string unit_string = txt.Substring(i).Trim().ToLower();
                if (unit_string.StartsWith("mi")) unit = Units.Miles;
                else if (unit_string.StartsWith("y")) unit = Units.Yards;
                else if (unit_string.StartsWith("f")) unit = Units.Feet;
                else if (unit_string.StartsWith("'")) unit = Units.Feet;
                else if (unit_string.StartsWith("k")) unit = Units.Kilometers;
                else if (unit_string.StartsWith("m")) unit = Units.Meters;
                else unit = Units.Undefined;
            }
        }

        // Return the length of the longest prefix
        // string that makes sense as a double.
        private int DoublePrefixLength(string txt)
        {
            for (int i = 1; i <= txt.Length; i++)
            {
                string test_string = txt.Substring(0, i);
                double test_value;
                if (!double.TryParse(test_string, out test_value)) return i - 1;
            }
            return txt.Length;
        }

        // Exit.
        private void mnuFileExit_Click(object sender, EventArgs e)
        {
            this.Close();
        }

        // Let the user draw something and calculate its length.
        private void btnDistance_Click(object sender, EventArgs e)
        {
            lblInstructions.Text = "Click and draw to define the path that you want to measure.";
            picMap.Cursor = Cursors.Cross;

            DistancePoints = new List<Point>();
            picMap.MouseDown += Distance_MouseDown;
        }

        private List<Point> DistancePoints;
        private void Distance_MouseDown(object sender, MouseEventArgs e)
        {
            DistancePoints.Add(e.Location);
            picMap.MouseDown -= Distance_MouseDown;
            picMap.MouseMove += Distance_MouseMove;
            picMap.MouseUp += Distance_MouseUp;
        }
        private void Distance_MouseMove(object sender, MouseEventArgs e)
        {
            DistancePoints.Add(e.Location);
            DisplayDistanceCurve();
        }
        private void Distance_MouseUp(object sender, MouseEventArgs e)
        {
            picMap.MouseMove -= Distance_MouseMove;
            picMap.MouseUp -= Distance_MouseUp;
            picMap.Cursor = Cursors.Default;
            lblInstructions.Text = "";

            // Measure the curve.
            double distance = 0;
            for (int i = 1; i < DistancePoints.Count; i++)
            {
                int dx = DistancePoints[i].X - DistancePoints[i - 1].X;
                int dy = DistancePoints[i].Y - DistancePoints[i - 1].Y;
                distance += Math.Sqrt(dx * dx + dy * dy);
            }

            // Convert into the proper units.
            CurrentDistance = distance * ScaleDistanceInUnits / ScaleDistanceInPixels;

            // Display the result.
            btnDistance.Text = string.Format("Distance: {0} {1}",
                CurrentDistance, CurrentUnit);
        }

        // Display the distance curve.
        private void DisplayDistanceCurve()
        {
            Bitmap bm = (Bitmap)Map.Clone();
            using (Graphics gr = Graphics.FromImage(bm))
            {
                gr.DrawLines(Pens.Red, DistancePoints.ToArray());
            }
            picMap.Image = bm;
        }
    }
}
